# Chapter 10: Toplevels

As previously discussed, toplevel surfaces represent the typical windows containing a title bar and buttons for close, minimize, and maximize actions. They are perhaps the most common type of surfaces encountered in a compositor during everyday use.

<center>
<img style="max-width:45vw" src="https://lh3.googleusercontent.com/PvNaxLgkjaPryJ8W_P3bDOccQu1m-zNEcI7aH_R8WggzylV5LQZtuzLTUSImThDI8IVsAI9DERF4cwvSqPAEAyjsihHuPCQlZAbvu33iMC2iXvpTswZC3RBNDyKm1YEWDnnKeCn2Qw=w2400"/>
</center>

Toplevels can either exist independently or serve as children of other toplevels, such as dialogs do. The functionality of toplevels is primarily managed within the Louvre::LToplevelRole class. In this chapter, we will learn how to interact with them and handle their various requests.

### State Flags

Toplevel surfaces can exhibit different states, which are listed in the Louvre::LToplevelRole::State flags enum. Among these states, the most common ones include:

* Louvre::LToplevelRole::Activated: When this flag is set, the toplevel's decorations are highlighted above others. Only one toplevel can be active at a time.

* Louvre::LToplevelRole::Maximized: When this flag is enabled, the toplevel typically removes its decorations and occupies a significant portion of the screen.

* Louvre::LToplevelRole::Fullscreen: Enabling this flag typically results in the toplevel removing its decorations and completely occupying the entire screen, always staying displayed above other surfaces (except for its child surfaces).

* Louvre::LToplevelRole::Resizing: This flag indicates an ongoing interactive resizing session, usually initiated when the user drags one of its corners.

> Toplevels can also request to be minimized, but this is not considered a toplevel state.

### Current and Pending States

To set the state flags of a toplevel, you can select from various Louvre::LToplevelRole::configure() variants, which also allow you to specify the desired size. If you provide a size of (0,0), the client is at liberty to determine its own dimensions. Retrieve the current toplevel state flags and size using the Louvre::LToplevelRole::states() and Louvre::LToplevelRole::size() properties.

It's crucial to recognize that the state of a toplevel doesn't immediately change upon calling Louvre::LToplevelRole::configure(). You must await the respective property listener to detect the actual state change. For instance, if you configure a toplevel with the Louvre::LToplevelRole::Maximized flag, wait for the Louvre::LToplevelRole::maximizedChanged() event to ensure proper positioning. Alternatively, monitor changes in any state using Louvre::LToplevelRole::statesChanged().

As mentioned earlier, you can access the current size and state flags of the toplevel with Louvre::LToplevelRole::size() and Louvre::LToplevelRole::states(), respectively. Meanwhile, Louvre::LToplevelRole::pendingSize() and Louvre::LToplevelRole::pendingStates() represent the values from the last Louvre::LToplevelRole::configure() that has not yet been acknowledged.

Once a configuration is acknowledged by the client (when the current state and size actually change), the values of pending state and size are updated to match the current state and size. It's important to note that you will likely call Louvre::LToplevelRole::configure() multiple times before these changes are acknowledged by the client. Therefore, using the pending size and state flags when configuring is advisable to avoid undesired results stemming from potential race conditions.

### Window Geometry

The size of a toplevel is not determined by its surface size but by its window geometry (`Louvre::LToplevelRole::windowGeometry().size()` or `Louvre::LToplevelRole::size()`). The window geometry specifies the rect within the surface that encompasses the actual toplevel window, excluding its decorative elements like shadows. This is the size you should take into account when configuring it.

<center>
<img style="max-width:25vw;min-width:400px" src="https://lh3.googleusercontent.com/pw/AIL4fc9WuCZOWGGC6v7ZepWA7VOWygycmNFEngEHX7J9ODdmzm1Qs4au1KYPMRJ_102tfvOXFBcL-Kwghm0bSKbd4U6der3p94l1drpn3ONqb9yfv-uFgHQ=w2400"/>
</center>

### Decorations

By default, clients handle their own toplevel decorations. However, some applications, like Qt applications, implement the [XDG Decoration](https://wayland.app/protocols/xdg-decoration-unstable-v1) protocol, allowing the compositor to take charge of drawing the decorations.

It's worth noting that this protocol is still somewhat experimental, and only a limited number of clients support it. Therefore, for the purpose of this tutorial, we will focus on client-side decorations. If you're interested in exploring server-side decorations, please refer to the [louvre-views](md_md__examples.html#views) example, which creates macOS-style decorations for toplevels that support it.

### Interactive Events

Clients can request to initiate interactive moving or resizing of toplevels, typically triggered by the user clicking on its top bar or one of its corners. We will explore how to handle these requests shortly.

Let's begin by crafting our own Louvre::LToplevelRole subclass, which we'll name `EToplevel`:

#### src/EToplevel.h

```cpp
#ifndef ETOPLEVEL_H
#define ETOPLEVEL_H

#include <LToplevelRole.h>

using namespace Louvre;

class ESurface;

class EToplevel : public LToplevelRole
{
public:
    EToplevel(Params *params);

    // Cast LSurface to ESurface
    inline ESurface *surf() const
    {
        return (ESurface*) surface();
    };

    void configureRequest() override;
    void setMinimizedRequest() override;
};

#endif // ETOPLEVEL_H
```

Here, we're overriding the Louvre::LToplevelRole::configureRequest() virtual method. This method is invoked when the toplevel is created and every time it's mapped after being previously unmapped. Within this method, we should call Louvre::LToplevelRole::configure() to properly set the state and size of the toplevel.

Additionally, we've overridden Louvre::LToplevelRole::setMinimizedRequest(), where we will allow the client to minimize the toplevel only if it's not in fullscreen mode. Finally, we've created an auxiliary method to perform the casting of Louvre::LSurface into our custom `ESurface` type.

#### src/EToplevel.cpp

```cpp
#include "EToplevel.h"
#include "ESurface.h"

EToplevel::EToplevel(Params *params) : LToplevelRole(params) {}

void EToplevel::configureRequest()
{
    // Request the client to draw its own window decorations
    setDecorationMode(ClientSide);

    // Activates the toplevel with size (0,0) so that the client can decide the size
    configure(LSize(0, 0), pendingStates() | LToplevelRole::Activated);
}

void EToplevel::setMinimizedRequest()
{
    if (!fullscreen())
        surf()->setMinimized(true);
}
```

In this case, we're requesting the client to use client-side decorations. Additionally, we're configuring the toplevel with a size of (0,0), allowing it to determine its own size. Furthermore, we're adding the Louvre::LToplevelRole::Activated state using the "|" operator. If you're not yet familiar with working with bit flags, you can refer to [this post](https://dietertack.medium.com/using-bit-flags-in-c-d39ec6e30f08) by Dieter Tack for a detailed guide on handling them.

Additionally, don't overlook its virtual constructor.

#### src/ECompositor.h

```cpp
    // ...

    // Virtual constructors
    LOutput *createOutputRequest() override;
    LSurface *createSurfaceRequest(LSurface::Params *params) override;
    LPointer *createPointerRequest(LPointer::Params *params) override;
    LKeyboard *createKeyboardRequest(LKeyboard::Params *params) override;
    LToplevelRole *createToplevelRoleRequest(LToplevelRole::Params *params) override;

    // ...
```

#### src/ECompositor.h

```cpp
// ...

#include "EToplevel.h"

// ...

LToplevelRole *ECompositor::createToplevelRoleRequest(LToplevelRole::Params *params)
{
    return new EToplevel(params);
}
```

After recompiling and running the compositor, you should observe that everything remains unchanged. To explore further, consider configuring the toplevels with different sizes and state flags to observe their responses.

You may have also noticed that upon creation, all toplevels are automatically centered on the screen. This behavior is a result of the default implementation of Louvre::LSurface::mappingChanged(), which centers them each time they are remapped.

To modify this behavior, let's override it and introduce a boolean variable in our `ESurface` class. This variable will ensure that the toplevels are centered only during their initial mapping.

#### src/ESurface.h

```cpp
// ...

class EToplevel;

class ESurface : public LSurface
{
public:

    // ...

    void mappingChanged() override;

    // ...

    bool firstMapping = true;

    // Cast LToplevelRole to EToplevel
    inline EToplevel *tl()
    {
        return (EToplevel*)toplevel();
    };
};
```

We initialize `firstMapping` as `true`, and after the initial mapping, we set it to `false`.

#### src/ESurface.cpp

```cpp
// ...

#include "EToplevel.h"

// ...

void ESurface::mappingChanged()
{
    if (mapped() && firstMapping)
    {
        firstMapping = false;

        // We only center toplevels the first time they are mapped
        if (tl())
        {
            // Use the output where the cursor is currently located
            LOutput *output = cursor()->output();

            // Restrict the area to the output size - the topbar area
            LSize availableSize = output->size() - LSize(0, TOPBAR_HEIGHT);

            // Center the toplevel
            setPos( (output->pos() + LPoint(0, TOPBAR_HEIGHT)) + (availableSize - tl()->windowGeometry().size()) / 2);

            // Prevent positioning it under the topbar (when the toplevel height is greater than dstSize.h())
            if (pos().y() < output->pos().y() + TOPBAR_HEIGHT)
                setY(output->pos().y() + TOPBAR_HEIGHT);

            output->repaint();
        }

        raise();
    }

    /* Note: LSurfaceViews are automatically hidden when their surfaces are unmapped */

    compositor()->repaintAllOutputs();
}

// ...
```

Louvre::LSurface::mappingChanged() is invoked whenever there's a change in the Louvre::LSurface::mapped() property. In contrast to the default approach, we have improved it by considering the topbar's height when centering the toplevel. When a toplevel's size exceeds the output size minus the topbar region, we reposition it downward to prevent it from appearing beneath the topbar.

### Interactive Toplevel Movement

Clients have the capability to request the initiation of interactive toplevel movement sessions, typically triggered when a user drags a toplevel window's titlebar. This request is managed within the Louvre::LToplevelRole::startMoveRequest() virtual method. Louvre also provides auxiliary methods for handling these sessions through the Louvre::LPointer class.

To initiate an interactive toplevel movement session, use Louvre::LPointer::startMovingToplevel(). This method accepts several arguments, including the target toplevel window you intend to move and the defined constraint area that specifies where the toplevel can be positioned. 

The toplevel's position is then updated by invoking Louvre::LPointer::updateMovingToplevelPos() within a pointer move event, as we previously demonstrated in [Chapter 8: Pointer Events](md_md_tutorial_08.html), taking the cursor's position into consideration. 

Lastly, to conclude the session, we call Louvre::LPointer::stopMovingToplevel() when releasing the left pointer button in a pointer button event as we also saw.

> You are not obligated to employ the Louvre::LPointer auxiliary methods for managing interactive moving toplevel sessions. Instead, you can manually update their positions using Louvre::LSurface::setPos(), as previously explained.

Let's customize the default behavior of Louvre::LToplevelRole::startMoveRequest() to prevent toplevel windows from being positioned under the topbar.

#### src/EToplevel.h

```cpp
// ...

class EToplevel : public LToplevelRole
{
public:
    
    // ...

    void startMoveRequest() override;
};

// ...
```

#### src/EToplevel.cpp

```cpp
#include <LSeat.h>
#include <LPointer.h>
#include <LCursor.h>
#include "ESurface.h"
#include "EToplevel.h"
#include "Global.h"

// ...

void EToplevel::startMoveRequest()
{
    if (!fullscreen() && !maximized() && surf()->hasPointerFocus())
        seat()->pointer()->startMovingToplevel(this, cursor()->pos(), LPointer::EdgeDisabled, TOPBAR_HEIGHT);
}
```

Here, we initiate the moving session only if the toplevel has pointer focus and is not in fullscreen mode or maximized. Additionally, we impose a constraint on the top edge position by assigning it the `TOPBAR_HEIGHT` macro as the value for the T (top) argument.

It's worth noting that, by default, all edges are disabled. You have the option to disable a specific edge restriction by using the Louvre::LPointer::EdgeDisabled value.

Upon recompiling and running the compositor again, you'll observe that toplevel windows can no longer be positioned under the topbar. However, if you resize them, they can still do it. Let's explore how to address this issue.

### Interactive Toplevel Resizing

Much like interactive movement, clients can initiate a resizing session by requesting it through Louvre::LToplevelRole::startResizeRequest(). This is commonly triggered when users drag one of the corners or edges of a toplevel window.

The Louvre::LPointer class also provides auxiliary methods to manage these toplevel resizing sessions. The process begins with a call to Louvre::LPointer::startResizingToplevel(). As the cursor position changes, we invoke Louvre::LPointer::updateResizingToplevelSize(), and each time the toplevel's size changes (handled within Louvre::LToplevelRole::geometryChanged()), we update its position using Louvre::LPointer::updateResizingToplevelPos(). Lastly, upon releasing the left pointer button (implemented in [Chapter 8: Pointer Events](md_md_tutorial_08.html)), we call Louvre::LPointer::stopResizingToplevel() to conclude the resizing session.

> It's important to note that both moving and resizing sessions will automatically stop if the toplevel is destroyed. Therefore, you don't need to be overly cautious about handling these situations yourself; the library takes care of this for you.

The Louvre::LPointer auxiliary methods internally configure the toplevel with the Louvre::LToplevelRole::Resizing flag during the session. If you want to handle toplevel resizing sessions manually, you should configure them with that flag and unset it when it finishes.

Let's customize the default behavior of Louvre::LToplevelRole::startResizeRequest() to prevent toplevel windows from being positioned under the topbar.

#### src/EToplevel.h

```cpp
// ...

class EToplevel : public LToplevelRole
{
public:
    
    // ...

    void startResizeRequest(ResizeEdge edge) override;
    void geometryChanged() override;
};

// ...
```

#### src/EToplevel.cpp

```cpp
// ...

void EToplevel::startResizeRequest(ResizeEdge edge)
{
    if (!fullscreen() && !maximized() && surf()->hasPointerFocus())
        seat()->pointer()->startResizingToplevel(this, edge, cursor()->pos(), LSize(0, 0), LPointer::EdgeDisabled, TOPBAR_HEIGHT);
}

void EToplevel::geometryChanged()
{
    if (resizing())
        updateResizingPos();
}
```

Just like with the interactive moving session, we ignore the request if the toplevel is in fullscreen mode, maximized or if its surface has no pointer focus.

As for Louvre::LPointer::startResizingToplevel(), we pass it the toplevel being resized, the edge or corner being dragged, the current cursor position, the minimum toplevel size `LSize(0,0)` and also the positioning constraints for each toplevel size. In this case `TOPBAR_HEIGHT` for the top edge to prevent it from being positioned under the topbar.\n

If you recompile and run it again, you will notice that the toplevels no longer position under the topbar while being resized. 

<img src="https://lh3.googleusercontent.com/pw/ADCreHe9h6_f3uF5llrKa4nt-n88dwoIDXtS-QA67EFT6PVFtDHdxlE5dP74X46eTuJ14sI6F_iEuQtBnsPUn0YjRHCqO0ENrLd3lBWM0d1o8lvbHUMINjY=w2400"/>

> You might also observe that **weston-terminal** leaves an empty space between its toplevel window and the top bar when being resized against it. This behavior is due to **weston-terminal** using fixed size steps and is unrelated to Louvre. If you try another client, such as **gedit**, you'll notice that this issue doesn't occur.

Let's now discuss how to handle maximization requests.

### Handling Maximization Requests

Clients can request to maximize their toplevels by using Louvre::LToplevelRole::setMaximizedRequest() and unmaximize them using Louvre::LToplevelRole::unsetMaximizedRequest(). This is typically done in response to user actions such as clicking the maximize button or pressing a keyboard shortcut.

Within these requests, we can configure the toplevels with or without the Louvre::LToplevelRole::Maximized flag and specify the desired size. We then wait for the Louvre::LToplevelRole::maximizedChanged() event to ensure proper positioning.

Let's override these methods to configure the toplevels based on the output size, subtracting the topbar region. Additionally, we will introduce auxiliary variables and methods to store and recover the toplevel's state to its previous "normal" state before maximizing it.

#### src/EToplevel.h

```cpp
// ...

class EOutput;

class EToplevel : public LToplevelRole
{
public:
    
    // ...

    void setMaximizedRequest() override;
    void unsetMaximizedRequest() override;
    void maximizedChanged() override;

    // Normal state
    void saveNormalState();
    void restoreNormalState();
    EOutput *normalOutput = nullptr;
    LPointF normalPos;
    LSize normalSize;
};

// ...
```

#### src/EToplevel.cpp

```cpp
// ...

#include "EOutput.h"
#include "ECompositor.h"

// ...

void EToplevel::geometryChanged()
{
    if (resizing())
        updateResizingPos();
    else if (maximized())
    {
        // Check if it properly configured to the size we specified
        if (normalOutput && (normalOutput->size() - LSize(0, TOPBAR_HEIGHT) == windowGeometry().size()))
            surf()->setPos(normalOutput->pos() + LSize(0, TOPBAR_HEIGHT));

        // If not, fallback to normal
        else
            configure(normalSize, Activated);
    }
}

void EToplevel::setMaximizedRequest()
{
    // Ignore the request if the surface has no pointer, keyboard focus or if it is child of a fullscreen toplevel
    if (maximized() || !(surf()->hasPointerFocus() || surf()->hasKeyboardFocus()) || surf()->view.parent() != &G::compositor()->surfacesLayer)
        return;

    // If the toplevel is not in fullscreen mode we save its current normal state
    if (!fullscreen())
        saveNormalState();

    // Check if normalOutput is not nullptr just for safety reasons
    if (normalOutput)
    {
        configure(normalOutput->size() - LSize(0, TOPBAR_HEIGHT), Activated | Maximized);
        return;
    }
}

void EToplevel::unsetMaximizedRequest()
{
    // Remove the Maximized flag and wait for maximizedChanged()
    if (maximized())
        configure(normalSize, Activated);
}

void EToplevel::maximizedChanged()
{
    // Fake a geometryChanged() event
    if (maximized())
    {
        geometryChanged();
        surf()->raise();
    }
    // Restore the normal state if changed from maximized to non fullscreen
    else
        if (!fullscreen())
            restoreNormalState();
}

/* Save the toplevel state before being maximized or set into fullscreen mode*/
void EToplevel::saveNormalState()
{
    normalOutput = surf()->primaryOutput();

    if (!normalOutput)
        normalOutput = (EOutput*)cursor()->output();

    normalSize = windowGeometry().size();
    normalPos = LSizeF(surf()->pos() - normalOutput->pos()) / normalOutput->size();
}

void EToplevel::restoreNormalState()
{
    if (!normalOutput)
        normalOutput = (EOutput*)cursor()->output();

    surf()->setPos(normalOutput->pos() + normalPos * normalOutput->size());
    surf()->sendOutputEnterEvent(normalOutput);

    if (surf()->view.parent() != &G::compositor()->surfacesLayer)
        G::moveSurfaceWithChildren(surf(), &G::compositor()->surfacesLayer);

    surf()->raise();
    normalOutput = nullptr;
}
```

In Louvre::LToplevelRole::setMaximizedRequest(), we configure the toplevel with the Louvre::LToplevelRole::Maximized flag, adjusting it to the primary output size minus the topbar region. We also call `saveNormalState()` to store the current primary output, as well as the toplevel's position and size for later restoration. We then await a configure ACK response through Louvre::LToplevelRole::maximizedChanged() event to properly position the toplevel. If the toplevel isn't configured with the specified size, we revert it to its previous state. We perform this check whenever the window geometry changes.

In Louvre::LToplevelRole::unsetMaximizedRequest(), we simply configure the toplevel with its previous size and state. When unmaximized, it restores its previous position using `restoreNormalState()`. After recompiling and running the compositor, you'll observe that toplevels no longer maximize themselves behind the topbar, and upon unmaximizing, they return to their previous position and size.

<img src="https://lh3.googleusercontent.com/pw/ADCreHdWo70LhuTMAzHGuj-JukFE9xgmTmf-yj0YUeuMXbLmFXoffBvJtb49yDDQUu9XxK_GmUm2Jj7ct68y0A9FoaPtddBSH0DfOhIIwMsLXSYXU44c3JE=w2400"/>

I made a drawing of you, successfully creating your own compositor with Louvre.

### Handling Fullscreen Requests

In order to display toplevels in fullscreen mode, we'll create a new black Louvre::LSolidColorView and relocate the toplevel, along with its child surfaces, into this view. This approach is necessary because toplevels have a defined Louvre::LToplevelRole::maxSize() that might be smaller than the output size. In such cases, it's important to center the toplevel on the screen to ensure a visually appealing appearance. If other surfaces were left in the background during fullscreen mode, it could appear somewhat unusual.

Just like with maximizing toplevels, we will store the current state before entering fullscreen mode and restore it when exiting fullscreen mode.

#### src/EToplevel.h

```cpp
// ...

#include <LSolidColorView.h>

// ...

class EToplevel : public LToplevelRole
{
public:
    
    // ...

    void setFullscreenRequest(LOutput *output) override;
    void unsetFullscreenRequest() override;
    void fullscreenChanged() override;
    LSolidColorView fullscreenBackground;

    // ...

};

// ...
```

#### src/EToplevel.cpp

```cpp
// ...

EToplevel::EToplevel(Params *params) :
    LToplevelRole(params),
    fullscreenBackground(0.f, 0.f, 0.f, 1.f, &G::compositor()->fullscreenLayer)
{
    fullscreenBackground.setVisible(false);
}

// ...

void EToplevel::geometryChanged()
{
    if (resizing())
        // ...
    else if (maximized())
    {
        // ...
    }
    else if (fullscreen())
    {
        // Check if it properly configured to the size we specified or smaller
        if (normalOutput && windowGeometry().w() <= normalOutput->size().w() && windowGeometry().h() <= normalOutput->size().h())
        {
            // Move the toplevel view and children to the fullscreenBackground layer
            if (surf()->view.parent() != &fullscreenBackground)
                G::moveSurfaceWithChildren(surf(), &fullscreenBackground);

            // Center the toplevel
            surf()->setPos(normalOutput->pos() + (normalOutput->size() - windowGeometry().size()) / 2);

            // Adjust the black background
            fullscreenBackground.setPos(normalOutput->pos());
            fullscreenBackground.setSize(normalOutput->size());

            // Bring the black background to the top
            fullscreenBackground.insertAfter(G::compositor()->fullscreenLayer.children().back());
            fullscreenBackground.setVisible(true);
            normalOutput->repaint();
        }
        // If not, fallback to normal
        else
            configure(normalSize, Activated);
    }
}

// ...

void EToplevel::setFullscreenRequest(LOutput *output)
{
    // Clients can specify in which output they want to fullscreen the toplevel,
    // but we will ignore it and use the one we select in saveNormalState().
    L_UNUSED(output);

    // Ignore the request if the surface has no pointer or keyboard focus or is already in fullscreen mode
    if (fullscreen() || !(surf()->hasPointerFocus() || surf()->hasKeyboardFocus()))
        return;

    // If the toplevel is not maximized we save its current normal state
    if (!maximized())
        saveNormalState();

    if (normalOutput)
    {
        configure(normalOutput->size(), Activated | Fullscreen);
        return;
    }
}

void EToplevel::unsetFullscreenRequest()
{
    // Remove the Fullscreen flag and wait for fullscreenChanged()
    if (fullscreen())
        configure(normalSize, Activated);
}

void EToplevel::fullscreenChanged()
{
    // Fake a geometryChanged() event
    if (fullscreen())
        geometryChanged();
    else
    {
        fullscreenBackground.setVisible(false);

        // Restore the normal state if changed from fullscreen to non maximized
        if (!maximized())
            restoreNormalState();
    }
}

// ...
```

In the constructor, we add the black fullscreen view to the fullscreen layer and initially hide it. Also, we don't need to explicitly define its translucent region as we did with the wallpaper texture view. Louvre::LScene automatically determines whether it is opaque or translucent based on its Louvre::LSolidColorView::opacity() value.

When a client requests to set a toplevel into fullscreen mode (Louvre::LToplevelRole::setFullscreenRequest()), we configure it with the size of the `normalOutput` chosen in `saveNormalState()`. The `output` param of Louvre::LToplevelRole::setFullscreenRequest() is the client preferred output for fullscreen mode, which can be `nullptr` if it doesn't prefer one. In this case we are simply ignoring it and using the one we selected.

Subsequently, when the toplevel's state transitions to fullscreen (Louvre::LToplevelRole::fullscreenChanged()), we conduct a size check against the output. If the toplevel's size exceeds that of the output, we revert it to the normal state. Conversely, if its size is smaller, we add it to the fullscreen background view, including all its child surfaces. In cases where the size is smaller, we center it as needed. We repeat this process whenever the toplevel's window geometry changes.

Upon a client's request to exit fullscreen mode, we return the toplevel to its previous 'normal' state.

> Adhering to the [XDG Shell protocol](https://wayland.app/protocols/xdg-shell#xdg_toplevel:request:set_maximized), we must always configure toplevels when clients request either setting or unsetting them in maximized or fullscreen mode. This applies even if for example a toplevel is already maximized, and the client seeks to maximize it again. If you do not explicitly configure the toplevels during these requests, Louvre will automatically handle the configuration for you, retaining the current pending states and size using `configure(pendingStates())`.

After recompiling and running the compositor, you'll observe toplevels seamlessly transitioning into fullscreen mode and returning to their normal state when exiting fullscreen mode. Clients usually let users toggle fullscreen mode by pressing `F1`.

### Managing Surface Parent Changes

As we relocate the fullscreen toplevels and their child surfaces into the black fullscreen view, it's essential to ensure that any newly created child surfaces are also relocated. To accomplish this, we can override the Louvre::LSurface::parentChanged() virtual method.

#### src/ESurface.h

```cpp
// ...

class ESurface : public LSurface
{
public:

    // ...

    void parentChanged() override;

    // ...
};

// ...
```

#### src/ESurface.cpp

```cpp
// ...

void ESurface::parentChanged()
{
    if (parent())
    {
        ESurface *p = (ESurface*)parent();
        view.setParent(p->view.parent());
    }
    else
        view.setParent(&G::compositor()->surfacesLayer);

    repaintOutputs();
}

// ...
```

So, whenever a surface changes its parent, we place it within the same parent view as its parent. If the surface has no parent, we position it within the surfaces layer.

### Managing Output Uninitialization or Rearrangement

In a manner similar to our approach for minimizing surfaces, we also need to address situations where the output stored in `normalOutput` becomes uninitialized while the toplevel is in a maximized or fullscreen state. Additionally, we haven't yet covered the scenario where surfaces are no longer visible within any output after such an event occurs. To handle these scenarios, we will implement code in Louvre::LOutput::uninitializeGL() and introduce a new method named `updateSurfaces()`.

#### src/EOutput.h

```cpp
// ...

class EOutput : public LOutput
{
public:

    // ...

    void updateSurfaces();
};

// ...
```

#### src/EOutput.cpp

```cpp
// ...

#include "EToplevel.h"

// ...

void EOutput::moveGL()
{
    // ...

    updateSurfaces();
    G::scene()->handleMoveGL(this);

    // ...
}

void EOutput::resizeGL()
{
    // ...

    updateSurfaces();
    G::scene()->handleResizeGL(this);
    
    // ...
}

// ...

void EOutput::uninitializeGL()
{
    // Set minimizedOutput and normalOutput pointers to nullptr to prevent the compositor from crashing
    for (ESurface *surface : G::surfaces())
    {
        if (surface->minimizedOutput == this)
            surface->minimizedOutput = nullptr;

        if (surface->tl())
        {
            if (surface->tl()->normalOutput == this)
            {
                surface->tl()->configure(surface->tl()->normalSize, LToplevelRole::Activated);
                surface->tl()->normalOutput = nullptr;
            }
        }
    }

    // ...
}

void EOutput::updateSurfaces()
{
    for (ESurface *surf : G::surfaces())
    {
        // Update maximized and fullscreen toplevels position and size
        if (surf->tl() && surf->tl()->normalOutput == this)
        {
            // Fake a geometryChanged() event
            surf->tl()->geometryChanged();
        }
        // Move surfaces that are not visible on any output into a random position within this output
        else
            if (!surf->primaryOutput())
                surf->setPos(pos() + LPoint(0, TOPBAR_HEIGHT) + LPoint(rand() % 300, rand() % 300));
    }
}

// ...
```

Now, when an output is uninitialized, we check toplevels with their `normalOutput` variable set to this one. In such cases, we set it to `nullptr` and return these toplevels to their normal state.

Finally, whenever there are changes in the size or position of the outputs, we trigger `updateSurfaces()`. Within this method, we simulate an Louvre::LToplevelRole::geometryChanged() event for toplevels that are either maximized or in fullscreen mode within the output. This ensures their proper positioning. Additionally, we reposition surfaces that are no longer visible to random locations within one of the still-initialized outputs to maintain their visibility to the user.

In the next chapter, we'll explore the handling of popup surfaces, which are notably simpler to manage than toplevels. See you there!

<a href="md_md_tutorial_09.html">◀ Chapter 9: Keyboard Events</a> || <a href="md_md_tutorial_11.html"> Chapter 11: Popups ▶</a>